from itertools import product
from typing import Dict, Any, List, Tuple
import json
import glob
import os
import copy
from ml_gym.io.config_parser import YAMLConfigLoader


class GridSearch:
    # GRID SEARCH CREATION
    @staticmethod
    def _get_dict_obj(keys: List[str], values: Tuple[Any]) -> Dict[str, Any]:
        """
        Merges two lists into a dictionary, where one acts as the keys and the other one as the values
        :param keys:
        :param values:
        :return:
        """
        d = {}
        for key, value in zip(keys, values):
            d[key] = value
        return d

    @staticmethod
    def _find_products(splits_by_keys: Dict[str, List[Any]]) -> List[Dict[str, Any]]:
        """
        Takes a dictionary str -> List, calculates the cartesian product for the set of lists and then assigns each product
        the respective key.
        Example: {"p1": ["A", "B", "C"], "p2": ["a", "b"], "p3": [1]} then the cartesian product will produce
        [("A", "a", 1), ("A", "b", 1), ("B", "a", 1), ("B", "b", 1), ("C", "a", 1), ("C", "b", 1)].
        From each tuple we then create a dictionary resulting in the list of dictionaries
        [{"p1": "A", "p2": "a", "p3": 1}, ..., {"p1": "C", "p2": "b", "p3": 1}].

        :param splits_by_keys:
        :return:
        """
        values = list(splits_by_keys.values())
        keys = list(splits_by_keys.keys())
        if len(values) == 1:
            dict_objs = [GridSearch._get_dict_obj(keys, (value,)) for value in values[0]]
        else:
            product_values = product(*values)
            dict_objs = [GridSearch._get_dict_obj(keys, value) for value in product_values]
        return dict_objs

    @staticmethod
    def _is_sweep_node(node: Dict[str, Any]):
        """
        Checks if a given node is a sweep node
        :param node:
        :return:
        """
        return GridSearch._is_node(node) and "sweep" in node

    @staticmethod
    def _is_node(item):
        """
        Checks if a given item is a node. Otherwise it's a leave.
        :param item:
        :return:
        """
        return isinstance(item, dict)

    @staticmethod
    def _expand_sweep_node(node: Dict[str, Any]) -> Dict[str, Any]:
        """
        Depending on the sweep type, e.g., range or absolute, the given sweep node needs to be fully created.
        :param node:
        :return:
        """
        if node["sweep"] == "absolute":
            return node
        else:
            raise NotImplementedError("Other sweep types have not been implemented, yet!")

    @staticmethod
    def _split_config(node: Dict[str, Any]) -> List[Dict[str, Any]]:
        """
        Takes a grid search dictionary or a node within a grid search and creates the corresponding list of configs.
        :param node:
        :return:
        """
        splits_by_key = {}
        for child_key, child in node.items():
            if GridSearch._is_sweep_node(child):  # sweep node
                child = GridSearch._expand_sweep_node(child)
                all_splits = []
                for item in child["values"]:  # for each element in the sweep
                    splits = GridSearch._split_config(item) if GridSearch._is_node(item) else [item]
                    all_splits.extend(splits)
                splits_by_key[child_key] = all_splits

            elif GridSearch._is_node(child):  # normal dictionary node
                splits_by_key[child_key] = GridSearch._split_config(child)
            else:  # leave node
                splits_by_key[child_key] = [child]

        configs = GridSearch._find_products(splits_by_key)
        return [copy.deepcopy(config) for config in configs]

    # GS creationu

    @staticmethod
    def create_gs_configs_from_path(config_path: str) -> List[Dict]:
        gs_config = YAMLConfigLoader.load(config_path)
        configs = GridSearch.create_gs(gs_config)
        return configs

    @staticmethod
    def create_gs(gs: Dict):
        configs = GridSearch._split_config(gs)
        return configs

    # CONFIG PART OF GS

    @staticmethod
    def is_config_in_gs(d: Dict, gs: Dict, negligible_paths: Dict = None) -> bool:
        if negligible_paths is None:
            negligible_paths = {}
        gs_configs = GridSearch._split_config(gs)
        for gs_config in gs_configs:
            if GridSearch._is_config_equal(d, gs_config, negligible_paths):
                return True
        return False

    @staticmethod
    def _delete_branches(d, negligible_paths):
        """
        negligible example:
        ```
        {
        "optimizer": {
           "lr": None,
            "weight_decays": [0.0001, 123],
            }
        }
        ```
        ---> sets path root->optimizer->lr in d to None.
        Args:
            d:
            negligible_paths:

        Returns:

        """
        if isinstance(negligible_paths, dict):
            for key in negligible_paths.keys():
                if negligible_paths[key] is not None:
                    GridSearch._delete_branches(d[key], negligible_paths[key])
                else:
                    if key in d:
                        d.pop(key)
        if isinstance(negligible_paths, list):
            for i, negligible_elem in enumerate(negligible_paths):
                if negligible_elem is None:
                    d.pop(i)
                else:
                    GridSearch._delete_branches(d[i], negligible_paths[i])

    @staticmethod
    def _is_config_equal(d1: Dict, d2: Dict, negligible_paths: Dict = None) -> bool:
        """
        NOTE: This only works with dictionaries that don't have sweeps...
        Args:
            d1:
            d2:
            negligible_paths:

        Returns:

        """

        def ordered(obj):
            if isinstance(obj, dict):
                return sorted((k, ordered(v)) for k, v in obj.items())
            if isinstance(obj, list):
                return sorted(ordered(x) for x in obj)
            else:
                return obj

        if negligible_paths is None:
            negligible_paths = {}
        d1, d2 = d1.copy(), d2.copy()
        GridSearch._delete_branches(d1, negligible_paths)
        GridSearch._delete_branches(d2, negligible_paths)
        return ordered(d1) == ordered(d2)

    # Update configs from GS

    @staticmethod
    def update_config_from_grid_search(old_config: Dict, gs: List[Dict], negligible_paths: Dict):
        for new_config in gs:
            if GridSearch._is_config_equal(old_config, new_config, negligible_paths):
                return new_config
        raise Exception("Config is not present in grid search!")

    @staticmethod
    def get_rerun_configs(old_configs: List[Dict], gs: Dict, negligible_paths: Dict) -> List[Dict[str, Any]]:
        gs_configs = GridSearch.create_gs(gs)
        return [GridSearch.update_config_from_grid_search(old_config, gs_configs, negligible_paths) for old_config in
                old_configs]

    @staticmethod
    def are_sweeps_identical(gs_config_path: str, experiment_folder: str) -> bool:
        with open(gs_config_path, "r") as f:
            gs_config = json.load(f)
        gs_configs_list = GridSearch.create_gs_configs_from_path(gs_config_path)
        config_paths = glob.glob(os.path.join(experiment_folder, "**/config.json"), recursive=True)
        if len(gs_configs_list) != len(config_paths):
            print(f"GS lengths do not match! (gs: {len(gs_configs_list)} vs experiments: {len(config_paths)})")
            return False
        else:
            configs = []
            for path in config_paths:
                with open(path, "r") as f:
                    config = json.load(f)
                    configs.append(config)
            equal_map = [GridSearch.is_config_in_gs(d=config, gs=gs_config) for config in configs]
            print(equal_map)
            return all(equal_map)


if __name__ == "__main__":
    test_gs = {
        "p_1": 1,
        "p_2": {
            "sweep": "absolute",
            "values": [
                2,
                3,
                4
            ]
        },
        "p_3": {
            "sweep": "absolute",
            "values": [
                {
                    "p_3.1": "XYZ"
                },
                {
                    "p_3.1": {
                        "sweep": "absolute",
                        "values": [
                            6,
                            7
                        ]
                    }
                }
            ]
        },
        "p_4": [12, 13]
    }
    configs = GridSearch._split_config(test_gs)
    print(f" Total configs found: {len(configs)}")
    for config in configs:
        print(config)

    print(GridSearch._is_config_equal(configs[0], configs[1]))
    c = configs[0]
    c_negligible = {'p_3': {'p_3.1': None}, 'p_1': None, 'p_4': [12, None]}
    GridSearch._delete_branches(c, c_negligible)
    print(c)
    print(GridSearch._is_config_equal(c, c))



from datetime import datetime
from ml_gym.gym.jobs import AbstractGymJob
from ml_gym.gym.gym import Gym
from ml_gym.blueprints.blue_prints import BluePrint
from typing import List, Type
from ml_gym.util.grid_search import GridSearch
from ml_gym.util.logger import QueuedLogging
from multiprocessing import Queue
from shutil import copyfile
import os

def create_gym(process_count: int, device_ids, log_std_to_file: bool) -> Gym:
    gym = Gym(process_count, device_ids=device_ids, log_std_to_file=log_std_to_file)
    return gym


def setup_logging_environment(log_dir_path: str):
    queue = Queue()
    QueuedLogging.start_logging(queue, log_dir_path)


def stop_logging_environment():
    QueuedLogging.stop_listener()


def create_blueprints(blue_print_class: Type[BluePrint],
                      run_mode: AbstractGymJob.Mode,
                      gs_config_path: str,
                      dashify_logging_path: str,
                      num_epochs: int) -> List[BluePrint]:

    run_mode = AbstractGymJob.Mode[run_mode]
    run_id_to_config_dict = {str(run_id): config for run_id, config in enumerate(GridSearch.create_gs_configs_from_path(gs_config_path))}
    config_tuples = list(run_id_to_config_dict.items())
    epochs = list(range(num_epochs))
    grid_search_id = datetime.now().strftime("%Y-%m-%d--%H-%M-%S")
    gs_logging_path = os.path.join(dashify_logging_path, grid_search_id)
    os.makedirs(gs_logging_path, exist_ok=True)
    copyfile(gs_config_path, os.path.join(gs_logging_path, os.path.basename(gs_config_path)))
    blueprints = []
    for run_id, config in config_tuples:
        blue_print = blue_print_class(grid_search_id=grid_search_id,
                                      run_id=run_id,
                                      epochs=epochs,
                                      run_mode=run_mode,
                                      config=config,
                                      dashify_logging_dir=dashify_logging_path)
        blueprints.append(blue_print)
    return blueprints



from ml_gym.batching.batch import InferenceResultBatch
from typing import List, Callable
from functools import partial


class BatchFilter:

    @staticmethod
    def _class_filter_selection_fun(inference_batch_result: InferenceResultBatch, selected_class: int,
                                    target_subscription_key: str) -> List[bool]:
        return (inference_batch_result.targets[target_subscription_key] == selected_class).tolist()

    @staticmethod
    def get_class_selection_fun(target_subscription_key: str, selected_class: int = None) -> Callable[[InferenceResultBatch], List[bool]]:
        sample_selection_fun = partial(BatchFilter._class_filter_selection_fun,
                                       selected_class=selected_class,
                                       target_subscription_key=target_subscription_key)
        return sample_selection_fun



class DatasetNotFoundError(Exception):
    """Exception raised when a dataset could not be found."""
    pass


class DatasetOutOfBoundsError(Exception):
    """Exception raised when an index >= len(Dataset) is used."""
    pass


class ModelNotTrainedError(Exception):
    """Exception raised when model state is requested that is training dependent"""
    pass


class DatasetFileCorruptError(Exception):
    """Thrown when integrity checks indicate that a given file is corrupt."""
    pass


class LossFunctionInitError(Exception):
    """Exception raised when a loss function cannot be initiliazed"""
    pass


class BatchStateError(Exception):
    """Exception raised when a Batch object does not contain the requested data"""
    pass


class ExperimentInfoMissing(Exception):
    """Exception raised when experiment info object is missing"""
    pass


class ComponentConstructionError(Exception):
    """Exception raised when an error occurred during component construction"""
    pass


class SingletonAlreadyInstantiatedError(Exception):
    """Exception raised when trying to get another instance from a singleton class"""
    pass



from abc import abstractmethod
import torch
from ml_gym.gym.stateful_components import StatefulComponent
from typing import Dict, Any


class Scaler(StatefulComponent):

    @abstractmethod
    def train(self, loss: torch.Tensor):
        raise NotImplementedError

    @abstractmethod
    def scale(self, tensor: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


class MeanScaler(Scaler):
    def __init__(self):
        self._mean = 1

    @property
    def mean(self) -> float:
        return self._mean

    @mean.setter
    def mean(self, value: float):
        self._mean = value

    def scale(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor / self._mean

    def train(self, loss: torch.Tensor):
        self._mean = torch.mean(loss).item()

    def get_state(self) -> Dict[str, Any]:
        state = super().get_state()
        state["mean"] = self.mean
        return state

    def set_state(self, state: Dict[str, Any]):
        super().set_state(state)
        self._mean = state["mean"]


class NoScaler(Scaler):
    def __init__(self):
        pass

    def scale(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    def train(self, loss: torch.Tensor):
        pass



from typing import Tuple, Any, Dict, List
from abc import ABC, abstractmethod
from data_stack.dataset.iterator import DatasetIteratorIF
from ml_gym.data_handling.postprocessors.feature_encoder import CategoricalEncoder, ContinuousEncoder, Encoder
import torch
import numpy as np
import tqdm


class PostProcessorIf(ABC):

    @abstractmethod
    def postprocess(self, sample: Tuple[Any]) -> Tuple[Any]:
        raise NotImplementedError


class FittablePostProcessorIf(PostProcessorIf):

    @abstractmethod
    def fit(self, iterator: DatasetIteratorIF):
        raise NotImplementedError


class LabelMapperPostProcessor(PostProcessorIf):

    class Mapping:
        def __init__(self, previous_labels: List[Any], new_label: Any):
            self.previous_labels: List[Any] = previous_labels
            self.new_label: Any = new_label

        def __contains__(self, label: Any):
            return label in self.previous_labels

    def __init__(self, mappings: List[Dict], target_position: int = 1, tag_position: int = 2):
        self.target_position = target_position
        self.tag_position = tag_position
        self.mappings: List[LabelMapperPostProcessor.Mapping] = [LabelMapperPostProcessor.Mapping(**mapping) for mapping in mappings]

    def postprocess(self, sample: Tuple[Any]) -> Tuple[Any]:
        for mapping in self.mappings:
            if sample[self.target_position] in mapping:
                # have to convert to list as tuples are immutable
                sample = list(sample)
                sample[self.target_position] = mapping.new_label
            if sample[self.tag_position] in mapping:
                # have to convert to list as tuples are immutable
                sample = list(sample)
                sample[self.tag_position] = mapping.new_label
        return tuple(sample)


class FeatureEncoderPostProcessor(FittablePostProcessorIf):

    def __init__(self, sample_position: int, feature_encoding_configs: Dict[str, List[Any]], custom_encoders: Dict[str, Encoder] = None, sequential=False):
        self.sample_position = sample_position
        self.feature_encoding_configs = feature_encoding_configs
        self.feature_encoder_mapping: Dict[str, Encoder] = {"categorical": CategoricalEncoder, "continuous": ContinuousEncoder}
        if custom_encoders is not None:
            self.feature_encoder_mapping = {**self.feature_encoder_mapping, **custom_encoders}
        self.encoders: Dict[int, Encoder] = {}
        self.sequential = sequential

    def fit(self, iterators: Dict[str, DatasetIteratorIF]):

        def fit_sequential(iterators: Dict[str, DatasetIteratorIF]):
            encoders = {}
            for config in tqdm.tqdm(self.feature_encoding_configs, desc="Encoding Progress"):
                encoder_class = self.feature_encoder_mapping[config["feature_type"]]
                feature_indices = config["feature_names"]
                iterator = iterators[config["train_split"]]
                for feature_indice in tqdm.tqdm(feature_indices, desc="Feature type encoding"):
                    values = np.array([row[self.sample_position][feature_indice] for row in iterator])
                    encoder = encoder_class()
                    encoder.fit(values=values)
                    encoders[feature_indice] = encoder
            return encoders

        def fit_parallel(iterators: Dict[str, DatasetIteratorIF]):
            encoders = {}
            for config in tqdm.tqdm(self.feature_encoding_configs, desc="Encoding Progress"):
                encoder_class = self.feature_encoder_mapping[config["feature_type"]]
                feature_indices = config["feature_names"]
                iterator = iterators[config["train_split"]]
                subset_encoders = [encoder_class() for feature_indice in feature_indices]
                values = np.stack([row[self.sample_position][feature_indices] for row in iterator])

                for i in range(len(subset_encoders)):
                    subset_encoders[i].fit(values=values[:, i])
                    encoders[feature_indices[i]] = subset_encoders[i]
            return encoders

        if self.sequential:
            encoders = fit_sequential(iterators)
        else:
            encoders = fit_parallel(iterators)
        # order the encoders by their keys, as of python 3.6 insertion order equals iteration order
        self.encoders = {name: encoder for name, encoder in sorted(encoders.items(), key=lambda x: x[0])}

    def postprocess(self, sample: Tuple[Any]) -> Tuple[Any]:
        if not self.encoders:
            return sample
        sample = list(sample)  # need to make this a list because index assignment is not possible with tuples
        sample_input: torch.Tensor = sample[self.sample_position]
        sample[self.sample_position] = torch.from_numpy(np.hstack([self.encoders[pos].transform(np.array([val])).flatten()
                                                                   if pos in self.encoders.keys() else val for pos, val in enumerate(sample_input)]))
        return tuple(sample)

    def get_output_pattern(self):
        rep = ""
        lower = 0
        for index, encoder in self.encoders.items():
            upper = lower + encoder.get_output_size()
            rep += f"{index} ({type(encoder)}): {lower}, {upper} \n"
        return rep


class OneHotEncodedTargetPostProcessor(PostProcessorIf):

    def __init__(self, target_vector_size: int, target_position: int = 1):
        self.target_vector_size = target_vector_size
        self.target_position = target_position

    def postprocess(self, sample: Tuple[Any]) -> Tuple[Any]:
        target_vector = torch.zeros(self.target_vector_size)
        target_vector[sample[self.target_position]] = 1
        sample_list = list(sample)
        sample_list[self.target_position] = target_vector
        return tuple(sample_list)



from typing import Callable, Dict, Any
import torch
from sklearn.metrics import roc_auc_score, average_precision_score
from ml_gym.batching.batch import InferenceResultBatch
from abc import ABC


def binary_auroc_score(y_true: torch.Tensor, y_pred: torch.Tensor, **params: Dict[str, Any]) -> float:
    score = roc_auc_score(y_true=y_true, y_score=y_pred, **params)
    return score


def binary_aupr_score(y_true: torch.Tensor, y_pred: torch.Tensor, **params: Dict[str, Any]) -> float:
    score = average_precision_score(y_true=y_true, y_score=y_pred, **params)
    return score


class MetricIF(ABC):
    def __call__(self, result_batch: InferenceResultBatch) -> Any:
        raise NotImplementedError


class Metric(MetricIF):

    def __init__(self, tag: str, identifier: str):
        self.tag = tag
        self.identifier = identifier


class PredictionMetric(Metric):

    def __init__(self, tag: str, identifier: str, target_subscription_key: str,
                 prediction_subscription_key: str, metric_fun: Callable, params: Dict[str, Any]):
        super().__init__(tag=tag,
                         identifier=identifier)
        self.target_subscription_key = target_subscription_key
        self.prediction_subscription_key = prediction_subscription_key
        self.metric_fun = metric_fun
        self.params = params

    def __call__(self, result_batch: InferenceResultBatch) -> float:
        y_true = result_batch.get_targets(self.target_subscription_key).cpu()
        y_pred = result_batch.get_predictions(
            self.prediction_subscription_key).cpu()
        return self.metric_fun(y_true=y_true, y_pred=y_pred, **self.params)
